import json
import logging
from os import makedirs, getcwd
from os.path import dirname, join

import click
import pandas as pd
from multiprocessing import Pool
from .shared_params import (
    handle_project_id,
    handle_folder_id,
    project_id_arg,
    sample_ids_arg,
    handle_multiple_sample_ids,
    handle_multiple_result_file_ids,
    use_common_state,
    flatten_list_of_els_and_files,
    yes_option,
    module_option,
    ignore_errors_option,
    folder_ids_arg,
)
from geoseeq.result.file_download import download_url
from geoseeq.utils import download_ftp
from geoseeq.id_constructors import (
    result_file_from_uuid,
    result_file_from_name,
)
from geoseeq.knex import GeoseeqNotFoundError
from .progress_bar import PBarManager
from .utils import convert_size
from geoseeq.constants import FASTQ_MODULE_NAMES
from geoseeq.result import ResultFile
from geoseeq.upload_download_manager import GeoSeeqDownloadManager
from geoseeq.file_system.filesystem_download import (
    ProjectOnFilesystem,
    FILE_STATUS_MODIFIED_REMOTE,
    FILE_STATUS_MODIFIED_LOCAL,
    FILE_STATUS_NEW_LOCAL,
    FILE_STATUS_NEW_REMOTE,
    FILE_STATUS_IS_LOCAL_STUB,
)


logger = logging.getLogger('geoseeq_api')


@click.group("project")
def cli_project():
    """Download data from GeoSeeq."""
    pass


@cli_project.command("clone")
@use_common_state
@click.option('--use-stubs/--full-files', default=True, help='Download full files or stubs')
@click.option('--target-dir', '-d', default=None, help='Directory to download the project to')
@project_id_arg
def cli_clone_project(state, use_stubs, target_dir, project_id):
    """Clone a project to the local filesystem.
    """
    knex = state.get_knex().set_auth_required()
    proj = handle_project_id(knex, project_id)
    logger.info(f"Found project \"{proj.name}\"")
    if target_dir is None:
        target_dir = proj.name

    project = ProjectOnFilesystem(proj, target_dir)
    project.download(use_stubs=use_stubs)


@cli_project.command("status")
@use_common_state
def cli_project_status(state):
    """Check the status of a project on the local filesystem.
    """
    project = ProjectOnFilesystem.from_path(getcwd(), recursive=True)

    objs_by_status = {
        FILE_STATUS_MODIFIED_LOCAL: [],
        FILE_STATUS_MODIFIED_REMOTE: [],
        FILE_STATUS_NEW_LOCAL: [],
        FILE_STATUS_NEW_REMOTE: [],
        FILE_STATUS_IS_LOCAL_STUB: [],
    }
    for obj_type, status, local_path, obj in project.list_abnormal_objects():
        objs_by_status[status].append((obj_type, local_path, obj))

    print(f"Project: {project.project.name}")
    for status, objs in objs_by_status.items():
        print(f"Status: {status}")
        for obj_type, local_path, obj in objs:
            if status in (FILE_STATUS_MODIFIED_LOCAL, FILE_STATUS_NEW_LOCAL):
                print(f"  {obj_type}: {project.path_from_project_root(local_path)} -> {obj}")
            else:
                print(f"  {obj_type}:  {obj} -> {project.path_from_project_root(local_path)}")


import logging

import click
from geoseeq.knex import DEFAULT_ENDPOINT
import geoseeq.file_system_cache
from geoseeq import Knex
from geoseeq.utils import load_auth_profile

logger = logging.getLogger('geoseeq_api')


class State(object):

    def __init__(self):
        self.api_token = None
        self.endpoint = DEFAULT_ENDPOINT
        self.outfile = None
        self.log_level = 20
        self._knex = None
        self.use_cache = True

    def get_knex(self):
        logger.setLevel(self.log_level)
        self._knex = Knex(self.endpoint)
        if self.api_token:
            self._knex.add_api_token(self.api_token)
        return self._knex
    



pass_state = click.make_pass_decorator(State, ensure=True)


def log_level_option(f):
    def callback(ctx, param, value):
        state = ctx.ensure_object(State)
        state.log_level = value
        return value
    return click.option('-l', '--log-level',
                        type=click.Choice(['CRITICAL', 'ERROR', 'WARN', 'INFO', 'DEBUG']),
                        default='WARN',
                        envvar='GEOSEEQ_CLI_LOG_LEVEL',
                        expose_value=False,
                        callback=callback)(f)


def api_token_option(f):
    def callback(ctx, param, value):
        state = ctx.ensure_object(State)
        if not state.api_token:
            state.api_token = str(value)
        return value
    return click.option('-a', '--api-token',
                        envvar='GEOSEEQ_API_TOKEN',
                        expose_value=False,
                        help='Your GEOSEEQ API token.',
                        callback=callback)(f)


def endpoint_option(f):
    def callback(ctx, param, value):
        state = ctx.ensure_object(State)
        if not state.endpoint:
            state.endpoint = str(value)
        return value
    return click.option('--endpoint',
                        default=DEFAULT_ENDPOINT,
                        envvar='GEOSEEQ_API_ENDPOINT',
                        expose_value=False,
                        help='The URL to use for GEOSEEQ.',
                        callback=callback)(f)

def profile_option(f):
    def callback(ctx, param, value):
        state = ctx.ensure_object(State)
        endpoint, token = None, None
        if value:
            endpoint, token = load_auth_profile(value)
        else:
            try:
                endpoint, token = load_auth_profile()  # load default profile, if it exists
            except KeyError:
                pass
        if endpoint and token:
            state.endpoint = endpoint
            state.api_token = token
        return value
    return click.option('-p', '--profile',
                        envvar='GEOSEEQ_API_PROFILE',
                        expose_value=False,
                        help='The config profile to use for GEOSEEQ.',
                        callback=callback)(f)

def outfile_option(f):
    def callback(ctx, param, value):
        state = ctx.ensure_object(State)
        state.outfile = value
        return value
    return click.option('-o', '--outfile',
                        default='-', type=click.File('w'),
                        expose_value=False,
                        help='Output file path',
                        callback=callback)(f)


def cache_option(f):
    def callback(ctx, param, value):
        state = ctx.ensure_object(State)
        state.use_cache = value
        geoseeq.file_system_cache.USE_GEOSEEQ_CACHE = value
        return value
    return click.option('--use-cache/--no-cache',
                        default=True,
                        expose_value=False,
                        help='Cache data from the geoseeq server.',
                        callback=callback)(f)


def common_options(f):
    f = outfile_option(f)
    f = log_level_option(f)
    f = endpoint_option(f)
    f = profile_option(f)
    f = cache_option(f)
    return f


def use_common_state(f):
    f = common_options(f)
    f = pass_state(f)
    return f


def is_uuid(name):
    chunks = name.split('-')
    return len(chunks) == 5



import logging
import urllib

import pandas as pd

from .pipeline import Pipeline
from .remote_object import RemoteObject
from .result import ProjectResultFolder
from .sample import Sample
from .utils import paginated_iterator

logger = logging.getLogger("geoseeq_api")


class Project(RemoteObject):
    remote_fields = [
        "uuid",
        "created_at",
        "name",
        "privacy_level",
        "description",
        "samples_count",
    ]
    optional_remote_fields = [
        "privacy_level",
        "samples_count",
    ]
    parent_field = "org"
    url_prefix = "sample_groups"

    def __init__(
        self,
        knex,
        org,
        name,
        metadata={},
        is_library=True,
        is_public=False,
        storage_provider="default",
        privacy_level=None,
    ):
        super().__init__(self)
        self.knex = knex
        self.org = org
        self.new_org = None
        self.name = name
        self.is_library = is_library
        self.is_public = is_public
        self._sample_cache = []
        self._deleted_sample_cache = []
        self._get_sample_cache = []
        self._get_result_cache = []
        self.metadata = metadata
        self.storage_provider = storage_provider
        self.privacy_level = privacy_level

    def change_org(self, org):
        self.new_org = org
        self._modified = True

    def get_post_data(self):
        data = {field: getattr(self, field) for field in self.remote_fields if hasattr(self, field)}
        data["organization"] = self.org.uuid
        data['description'] = self.description if hasattr(self, 'description') and self.description else self.name
        data['privacy_level'] = self.privacy_level if hasattr(self, 'privacy_level') and self.privacy_level else 'private'
        data['storage_provider_name'] = self.storage_provider
        if self.new_org:
            if isinstance(self.new_org, RemoteObject):
                data["organization"] = self.new_org.uuid
            else:
                data["organization"] = self.new_org
        if data["uuid"] is None:
            data.pop("uuid")
        return data

    def nested_url(self):
        escaped_name = urllib.parse.quote(self.name, safe="")
        return self.org.nested_url() + f"/sample_groups/{escaped_name}"

    def _save_group_obj(self):
        data = self.get_post_data()
        url = f"sample_groups/{self.uuid}"
        self.knex.put(url, json=data)

    def _save_sample_list(self):
        sample_uuids = []
        for sample_uuid in self._sample_cache:
            sample_uuids.append(sample_uuid)
        if sample_uuids:
            url = f"sample_groups/{self.uuid}/samples"
            chunk_size = 100
            for i in range(0, len(sample_uuids), chunk_size):
                self.knex.post(url, json={"sample_uuids": sample_uuids[i : i + chunk_size]})
        self._sample_cache = []

    def _delete_sample_list(self):
        sample_uuids = []
        for sample_uuid in self._deleted_sample_cache:
            sample_uuids.append(sample_uuid)
        if sample_uuids:
            url = f"sample_groups/{self.uuid}/samples"
            self.knex.delete(url, json={"sample_uuids": sample_uuids})
        self._deleted_sample_cache = []

    def _save(self):
        self._save_group_obj()
        self._save_sample_list()
        self._delete_sample_list()

    def _get(self, allow_overwrite=False):
        """Fetch the result from the server."""
        self.org.idem()
        blob = self.get_cached_blob()
        if not blob:
            blob = self.knex.get(self.nested_url())
            self.load_blob(blob, allow_overwrite=allow_overwrite)
            self.cache_blob(blob)
        else:
            self.load_blob(blob)

    def _create(self):
        self.org.idem()
        post_data = self.get_post_data()
        blob = self.knex.post(
            f"sample_groups?format=json",
            json=post_data,
        )
        self.load_blob(blob)
    
    def add_sample_uuids(self, sample_uuids):
        """Return this group and add a sample to this group.

        Do not contact server until `.save()` is called on this group.
        """
        for sample_uuid in sample_uuids:
            self._sample_cache.append(sample_uuid)
        self._modified = True
        return self

    def add_sample(self, sample):
        """Return this group and add a sample to this group.

        Do not contact server until `.save()` is called on this group.
        """
        self._sample_cache.append(sample.uuid)
        self._modified = True
        return self

    def remove_sample(self, sample):
        """Return this group and remove a sample to this group.

        Do not contact server until `.save()` is called on this group.
        """
        self._deleted_sample_cache.append(sample.uuid)
        self._modified = True
        return self

    def sample(self, sample_name, metadata={}):
        return Sample(self.knex, self, sample_name, metadata=metadata)

    def result_folder(self, module_name, replicate=None, metadata={}):
        """Return a ProjectResultFolder object for this project."""
        return ProjectResultFolder(
            self.knex, self, module_name, replicate=replicate, metadata=metadata
        )

    def analysis_result(self, *args, **kwargs):
        """Return a ProjectResultFolder object for this project.
        
        Alias for result_folder."""
        return self.result_folder(*args, **kwargs)

    def get_samples(self, cache=True, error_handler=None):
        """Yield samples fetched from the server."""
        if cache and self._get_sample_cache:
            for sample in self._get_sample_cache:
                yield sample
            return
        url = f"sample_groups/{self.uuid}/samples"
        for sample_blob in paginated_iterator(self.knex, url, error_handler=error_handler):
            sample = self.sample(sample_blob["name"])
            sample.load_blob(sample_blob)
            sample.cache_blob(sample_blob)
            # We just fetched from the server so we change the RemoteObject
            # meta properties to reflect that
            sample._already_fetched = True
            sample._modified = False
            if cache:
                self._get_sample_cache.append(sample)
            else:
                yield sample
        if cache:
            for sample in self._get_sample_cache:
                yield sample

    def get_sample_uuids(self, cache=True, error_handler=None):
        """Yield samples uuids fetched from the server."""
        if cache and self._get_sample_cache:
            for sample in self._get_sample_cache:
                yield sample.uuid
            return
        url = f"sample_groups/{self.uuid}/samples"
        for sample_blob in paginated_iterator(self.knex, url, error_handler=error_handler):
            yield sample_blob['uuid']

    def _batch_sample_uuids(self, batch_size, input_sample_uuids=[]):
        """Yield batches of sample uuids."""
        uuids_to_batch = input_sample_uuids if input_sample_uuids else self.get_sample_uuids()
        sample_uuids = []
        for sample_uuid in uuids_to_batch:
            sample_uuids.append(sample_uuid)
            if len(sample_uuids) == batch_size:
                yield sample_uuids
                sample_uuids = []
        if sample_uuids:
            yield sample_uuids

    def get_analysis_results(self, cache=True):
        """Yield ProjectResultFolder objects for this project fetched from the server.
        
        Alias for get_result_folders."""
        return self.get_result_folders(cache=cache)

    def get_result_folders(self, cache=True):
        """Yield ProjectResultFolder objects for this project fetched from the server."""
        if cache and self._get_result_cache:
            for ar in self._get_result_cache:
                yield ar
            return
        url = f"sample_group_ars?sample_group_id={self.uuid}"
        result = self.knex.get(url)
        for result_blob in result["results"]:
            result = self.analysis_result(result_blob["module_name"])
            result.load_blob(result_blob)
            # We just fetched from the server so we change the RemoteObject
            # meta properties to reflect that
            result._already_fetched = True
            result._modified = False
            if cache:
                self._get_result_cache.append(result)
            else:
                yield result
        if cache:
            for ar in self._get_result_cache:
                yield ar

    def get_manifest(self):
        """Return a manifest for this group."""
        url = f"sample_groups/{self.uuid}/manifest"
        return self.knex.get(url)

    def get_module_counts(self):
        """Return a dictionary with module counts for samples in this group."""
        url = f"sample_groups/{self.uuid}/module_counts"
        return self.knex.get(url)

    def get_sample_metadata(self):
        """Return a pandas dataframe with sample metadata."""
        url = f"sample_groups/{self.uuid}/samples-list?page=1&page_size=500&&"
        rows = []
        while url:
            blob = self.knex.get(url)
            rows.extend(blob["results"])
            url = blob["next"]
        return pd.DataFrame(rows)
    
    
    @property
    def n_samples(self):
        """Return the number of samples in this project."""
        if hasattr(self, 'samples_count') and self.samples_count is not None:
            return self.samples_count
        return len(list(self.get_sample_uuids()))
    
    def bulk_find_files(self,
                        sample_uuids=[],
                        sample_name_includes=[],
                        folder_types="all",
                        folder_names=[],
                        file_names=[],
                        extensions=[],
                        with_versions=False,
                        use_batches_cutoff=500):
        """Return a dict with links to download files that match the given criteria.

        Options:
        - sample_uuids: list of sample uuids; if blank search all samples in project
        - sample_name_includes: list of strings; finds samples with names that include these strings
        - folder_types: "all", "project", "sample"; finds files in folders of these types
        - folder_names: list of strings; finds files in folders that have these strings in their names
        - file_names: list of strings; finds files that have these strings in their names
        - extensions: list of strings; finds files with these file extensions
        - with_versions: bool; if True, include all versions of files in results
        """
        def _my_bulk_find(sample_uuids=None):  # curry to save typing
            return self._bulk_find_files_batch(sample_uuids=sample_uuids or [],
                                             sample_name_includes=sample_name_includes,
                                             folder_types=folder_types,
                                             folder_names=folder_names,
                                             file_names=file_names,
                                             extensions=extensions,
                                             with_versions=with_versions)
        n_samples = len(sample_uuids) if sample_uuids else self.n_samples
        if n_samples < use_batches_cutoff:
            logger.debug(f"Using single batch bulk_find for {n_samples} samples")
            return _my_bulk_find(sample_uuids=sample_uuids)
        else:
            logger.debug(f"Using multi batch bulk_find for {n_samples} samples")
            merged_response = {'file_size_bytes': 0, 'links': {}, 'no_size_info_count': 0}
            for batch in self._batch_sample_uuids(use_batches_cutoff - 1, input_sample_uuids=sample_uuids):
                response = _my_bulk_find(sample_uuids=batch)
                merged_response['file_size_bytes'] += response['file_size_bytes']
                merged_response['links'].update(response['links'])
                merged_response['no_size_info_count'] += response['no_size_info_count']
            return merged_response
                
    def _bulk_find_files_batch(self,
                               sample_uuids=None,
                               sample_name_includes=None,
                               folder_types=None,
                               folder_names=None,
                               file_names=None,
                               extensions=None,
                               with_versions=False):
        data = {
            "sample_uuids": sample_uuids or [],
            "sample_names": sample_name_includes or [],
            "folder_type": folder_types or [],
            "folder_names": folder_names or [],
            "file_names": file_names or [],
            "extensions": extensions or [],
            "with_versions": with_versions
        }
        url = f"sample_groups/{self.uuid}/download"
        response = self.knex.post(url, data)
        return response
    
    def run_app(self, app: Pipeline, input_parameters=None):
        """Run an app on this group."""
        if not input_parameters:
            input_parameters = app.get_input_parameters()
        params = {
            'pipeline_id': app.uuid,
            'sample_uuids': list(self.get_sample_uuids()),
            'input_parameters': input_parameters,
        }
        self.knex.post(f"sample_groups/{self.uuid}/run_app", json=params, json_response=False)

    def __str__(self):
        return f"<Geoseeq::Project {self.name} {self.uuid} />"

    def __repr__(self):
        return f"<Geoseeq::Project {self.name} {self.uuid} />"

    def pre_hash(self):
        return "PROJ" + self.name + self.org.pre_hash()
    
    @property
    def grn(self):
        return f"grn::project:{self.uuid}"

SampleGroup = Project  # alias for backwards compatibility


import logging
from multiprocessing import Pool, current_process
from os.path import basename, join, dirname
from geoseeq.result import ResultFile
from geoseeq.result.file_download import download_url
from os import makedirs

logger = logging.getLogger('geoseeq_api')
logger.addHandler(logging.NullHandler())  # No output unless configured by calling program


def _make_in_process_logger(log_level):
    logger = logging.getLogger('geoseeq_api')
    logger.setLevel(log_level)
    handler = logging.StreamHandler()
    handler.setFormatter(logging.Formatter('[%(levelname)s] %(name)s :: ' + current_process().name + ' :: %(message)s'))
    logger.addHandler(handler)
    return logger


def _upload_one_file(args):
    (result_file, filepath, session, progress_tracker,
     link_type, overwrite, log_level, parallel_uploads,
     use_cache, no_new_versions, threads_per_upload,
     num_retries, ignore_errors, chunk_size_mb, use_atomic_upload) = args
    chunk_size = chunk_size_mb * 1024 * 1024 if chunk_size_mb else None
    if parallel_uploads:
        _make_in_process_logger(log_level)
    try:
        if link_type == 'upload':
            # TODO: check checksums to see if the file is the same
            result_file.upload_file(
                filepath,
                session=session, overwrite=overwrite, progress_tracker=progress_tracker,
                threads=threads_per_upload, use_cache=use_cache, chunk_size=chunk_size,
                no_new_versions=no_new_versions, max_retries=num_retries,
                use_atomic_upload=use_atomic_upload
            )
        else:
            result_file.link_file(link_type, filepath)
    except Exception as e:
        if ignore_errors:
            logger.error(f"Error uploading {filepath}: {e}")
        else:
            raise e
    return result_file


class GeoSeeqUploadManager:

    def __init__(self,
                 n_parallel_uploads=1,
                 threads_per_upload=4,
                 session=None,
                 link_type='upload',
                 progress_tracker_factory=None,
                 log_level=logging.WARNING,
                 overwrite=True,
                 no_new_versions=False,
                 num_retries=3,
                 ignore_errors=False,
                 chunk_size_mb=5,
                 use_atomic_upload=True,
                 use_cache=True):
        self.session = session
        self.n_parallel_uploads = n_parallel_uploads
        self.progress_tracker_factory = progress_tracker_factory if progress_tracker_factory else lambda x: None
        self.log_level = log_level
        self.link_type = link_type
        self.overwrite = overwrite
        self._result_files = []
        self.no_new_versions = no_new_versions
        self.use_cache = use_cache
        self.threads_per_upload = threads_per_upload
        self.num_retries = num_retries
        self.ignore_errors = ignore_errors
        self.chunk_size_mb = chunk_size_mb
        self.use_atomic_upload = use_atomic_upload

    def add_result_file(self, result_file, local_path):
        self._result_files.append((result_file, local_path))

    def add_local_file_to_result_folder(self, result_folder, local_path, geoseeq_file_name=None):
        if not geoseeq_file_name:
            if local_path.startswith("/"):  # if local path is an absolute path use the basename
                geoseeq_file_name = basename(local_path)
            else:
                # remove "./" and "../" from local path to get a geoseeq file name
                geoseeq_file_name = local_path.replace("./", "").replace("../", "")
        result_file = result_folder.result_file(geoseeq_file_name)
        self.add_result_file(result_file, local_path)

    def add_local_folder_to_result_folder(self, result_folder, local_path, recursive=False, hidden_files=False, prefix=""):
        for result_file, local_path in result_folder._prepare_folder_upload(local_path, recursive, hidden_files, prefix):
            self.add_result_file(result_file, local_path)

    def get_preview_string(self):
        out = ["Upload Preview:"]
        for result_file, local_path in self._result_files:
            out.append(f"{local_path} -> {result_file}")
        return "\n".join(out)

    def upload_files(self):
        upload_args = [(
                result_file, local_path,
                self.session, self.progress_tracker_factory(local_path),
                self.link_type, self.overwrite, self.log_level,
                self.n_parallel_uploads > 1, self.use_cache, self.no_new_versions,
                self.threads_per_upload, self.num_retries, self.ignore_errors,
                self.chunk_size_mb, self.use_atomic_upload
            ) for result_file, local_path in self._result_files
        ]
        out = []
        if self.n_parallel_uploads == 1:
            logger.info(f"Uploading files in series.")
            for upload_arg in upload_args:
                out.append(_upload_one_file(upload_arg))
        else:
            logger.info(f"Uploading files in parallel with {self.n_parallel_uploads} threads.")
            with Pool(self.n_parallel_uploads) as p:
                for uploaded_result_file in p.imap_unordered(_upload_one_file, upload_args):
                    out.append(uploaded_result_file)
        return out
    
    def __len__(self):
        return len(self._result_files)


def _download_one_file(args):
    url, file_path, pbar, ignore_errors, head, log_level, parallel_downloads = args
    if parallel_downloads:
        _make_in_process_logger(log_level)
    if isinstance(url, ResultFile):
        url = url.get_download_url()
    try:
        if dirname(file_path):
            makedirs(dirname(file_path), exist_ok=True)
        return download_url(url, filename=file_path, progress_tracker=pbar, head=head)
    except Exception as e:
        if ignore_errors:
            logger.error(f"Error downloading {url}: {e}")
        else:
            raise e
        

class GeoSeeqDownloadManager:

    def __init__(self, n_parallel_downloads=1, ignore_errors=False, head=False, progress_tracker_factory=None, log_level=logging.WARNING):
        self.n_parallel_downloads = n_parallel_downloads
        self.ignore_errors = ignore_errors
        self.head = head
        self.progress_tracker_factory = progress_tracker_factory if progress_tracker_factory else lambda x: None
        self.log_level = log_level
        self._result_files = []

    def add_download(self, url, file_path=None, progress_tracker=None):
        if not file_path:
            if isinstance(url, ResultFile):
                file_path = url.get_local_filename()
            else:
                raise ValueError("file_path must be provided if url is not a ResultFile object.")
        self._result_files.append((url, file_path))


    def add_result_folder_download(self, result_folder, local_folder_path, hidden_files=True):
        for result_file in result_folder.get_result_files():
            if not hidden_files and result_file.name.startswith("."):
                continue
            self.add_download(result_file, join(local_folder_path, result_file.get_local_filename()))

    def get_preview_string(self):
        out = ["Download Preview:"]
        for url, file_path in self._result_files:
            out.append(f"{url} -> {file_path}")
        return "\n".join(out)
    
    def get_url_string(self):
        self._convert_result_files_to_urls()
        out = []
        for url, _ in self._result_files:
            out.append(url)
        return "\n".join(out)
    
    def __len__(self):
        return len(self._result_files)
    
    def _convert_result_files_to_urls(self):
        self._result_files = [(
            url.get_download_url() if isinstance(url, ResultFile) else url,
            file_path,
        ) for url, file_path in self._result_files]

    def download_files(self):
        self._convert_result_files_to_urls()
        download_args = [(
            url, file_path,
            self.progress_tracker_factory(file_path),
            self.ignore_errors, self.head, self.log_level,
            self.n_parallel_downloads > 1
        ) for url, file_path in self._result_files]
        out = []
        if self.n_parallel_downloads == 1:
            logger.info(f"Downloading files in series.")
            for download_arg in download_args:
                out.append(_download_one_file(download_arg))
        else:
            logger.info(f"Downloading files in parallel with {self.n_parallel_downloads} threads.")
            with Pool(self.n_parallel_downloads) as p:
                for downloaded_file in p.imap_unordered(_download_one_file, download_args):
                    out.append(downloaded_file)
        return out



import json
import logging
import os
import time
import urllib.request
import urllib.parse
from os.path import basename, getsize, join
from pathlib import Path
from tempfile import NamedTemporaryFile

import requests

from geoseeq.constants import FIVE_MB
from geoseeq.remote_object import RemoteObject, RemoteObjectError
from geoseeq.utils import download_ftp, md5_checksum
from geoseeq.knex import GeoseeqOtherError

from .utils import *
from .file_upload import ResultFileUpload
from .file_download import ResultFileDownload


class ResultFile(RemoteObject, ResultFileUpload, ResultFileDownload):
    remote_fields = [
        "uuid",
        "created_at",
        "updated_at",
        "name",
        "stored_data",
        "pipeline_run",
    ]
    optional_remote_fields = [
        "pipeline_run",
    ]
    parent_field = "parent"

    def __init__(self, knex, parent, field_name, pipeline_run=None, data={}):
        super().__init__(self)
        self.knex = knex
        self.parent = parent
        self.name = field_name
        self.stored_data = data
        self.pipeline_run = pipeline_run
        self._cached_filename = None  # Used if the field points to S3, FTP, etc
        self._temp_filename = False

    @property
    def is_sample_result(self):
        return isinstance(self, SampleAnalysisResultField)

    @property
    def brn(self):
        obj_type = "sample" if self.canon_url() == "sample_ar_fields" else "project"
        brn = f"brn:{self.knex.instance_code()}:{obj_type}_result_field:{self.uuid}"

    def has_downloadable_file(self):
        """Return True if this field has a downloadable file."""
        try:
            self.download(head=10, cache=False)
            return True
        except Exception as e:
            return False

    def nested_url(self):
        escaped_name = urllib.parse.quote(self.name, safe="")
        return self.parent.nested_url() + f"/fields/{escaped_name}"

    def get_blob_filename(self):
        sname = self.parent.parent.name.replace(".", "-")
        mname = self.parent.module_name.replace(".", "-")
        fname = self.name.replace(".", "-")
        filename = join(self.parent.parent.name, f"{sname}.{mname}.{fname}.json").replace(
            "::", "__"
        )
        return filename

    def get_referenced_filename_ext(self):
        try:
            key = [k for k in ["filename", "uri", "url"] if k in self.stored_data][0]
        except IndexError:
            raise TypeError("Cannot make a reference filename for a BLOB type result field.")
        ext = self.stored_data[key].split(".")[-1]
        if ext in ["gz"]:
            ext = self.stored_data[key].split(".")[-2] + "." + ext
        return ext

    def get_referenced_filename(self):
        ext = self.get_referenced_filename_ext()
        sname = self.parent.parent.name.replace(".", "-").replace(" ", "_").lower()
        mname = self.parent.module_name.replace(".", "-").replace(" ", "_").lower()
        fname = self.name.replace(".", "-").replace(" ", "_").lower()
        filename = f"{sname}.{mname}.{fname}.{ext}".replace(
            "::", "__"
        )
        return filename

    def get_local_filename(self):
        """Return a filename that can be used to store this field locally."""
        return self.name
        # try:
        #     return basename(self.get_referenced_filename())
        # except TypeError:
        #     return basename(self.get_blob_filename())

    def _save(self):
        data = {field: getattr(self, field) for field in self.remote_fields if hasattr(self, field)}
        data["analysis_result"] = self.parent.uuid
        url = f"{self.canon_url()}/{self.uuid}"
        self.knex.put(url, json=data)

    def _get(self, allow_overwrite=False):
        """Fetch the result from the server."""
        self.parent.idem()
        blob = self.knex.get(self.nested_url())
        self.load_blob(blob, allow_overwrite=allow_overwrite)

    def _get_from_list(self, allow_overwrite=False):
        """Fetch the result from the server by listing the parent's children and finding this field.
        
        This function is a workaround for files with names that have funky characters that don't 
        work well in URLs.
        """
        self.parent.idem()
        for field in self.parent.get_result_files():
            if field.name == self.name:
                print(field)
                print()
                self.load_blob(field.get_blob(), allow_overwrite=allow_overwrite)
                self._already_fetched = True
                self._modified = False
                return

    def get_post_data(self):
        """Return a dict that can be used to POST this field to the server."""
        data = {
            "analysis_result": self.parent.uuid,
            "name": self.name,
            "stored_data": self.stored_data,
            "pipeline_run": self.pipeline_run,
        }
        return data

    def link_file(self, link_type, *args, **kwargs):
        if link_type == "s3":
            return self.link_s3(*args, **kwargs)
        elif link_type == "ftp":
            return self.link_ftp(*args, **kwargs)
        elif link_type == "sra":
            return self.link_sra(*args, **kwargs)
        elif link_type == "azure":
            return self.link_azure(*args, **kwargs)
        elif link_type == "http":
            return self.link_http(*args, **kwargs)
        assert False, f"Unknown link type: {link_type}"

    def link_s3(self, url, endpoint_url=None):
        """Link this field to an S3 object.

        Args:
            url (str): The URL of the S3 object.
            endpoint_url (str): The URL of the S3 endpoint.
        """
        if endpoint_url is None:
            if not url.startswith("https://"):
                raise ValueError("endpoint_url must be specified for non-HTTPS URLs.")
            endpoint_url = "https://" + url.split("https://")[1].split("/")[0]
            url = "s3://" + url.split(endpoint_url)[1][1:]
        self.stored_data = {
            "__type__": "s3",
            "uri": url,
            "endpoint_url": endpoint_url,
        }
        return self.save()

    def link_ftp(self, url):
        """Link this field to an FTP object.

        Args:
            url (str): The URL of the FTP object.
        """
        self.stored_data = {
            "__type__": "ftp",
            "uri": url,
        }
        return self.save()

    def link_sra(self, url):
        """Link this field to an SRA object.

        Args:
            url (str): The URL of the SRA object.
        """
        self.stored_data = {
            "__type__": "sra",
            "uri": url,
        }
        return self.save()

    def link_azure(self, url):
        """Link this field to an Azure object.

        Args:
            url (str): The URL of the Azure object.
        """
        endpoint_url = "https://" + url.split("https://")[1].split("/")[0]
        self.stored_data = {
            "__type__": "azure",
            "uri": url,
            "endpoint_url": endpoint_url,
        }
        return self.save()
    
    def link_http(self, url):
        """Link this field to an HTTP object.

        Args:
            url (str): The URL of the HTTP object.
        """
        self.stored_data = {
            "__type__": "http",
            "url": url,
        }
        return self.save()
    
    def idem(self):
        try:
            return super().idem()
        except GeoseeqOtherError as e:
            if "The fields analysis_result, name must make a unique set." in str(e):
                # this typically happens when the field name has a character that doesn't work well in URLs
                self._get_from_list(allow_overwrite=True)
            else:
                raise e
        return self

    def _create(self):
        check_json_serialization(self.stored_data)
        self.parent.idem()
        data = self.get_post_data()
        blob = self.knex.post(f"{self.canon_url()}?format=json", json=data)
        self.load_blob(blob)

    def __del__(self):
        if self._temp_filename and self._cached_filename:
            os.remove(self._cached_filename)

    def pre_hash(self):
        return self.name + self.parent.pre_hash()

    def copy(self, new_parent, save=True):
        copied = new_parent.field(self.name, data=self.stored_data)
        if save:
            copied.idem()
        return copied

    def checksum(self):
        """Return a checksum for this field as a blob.

        TODO
        """
        return {"value": "", "method": "none"}

AnalysisResultField = ResultFile

class SampleResultFile(ResultFile):
    def canon_url(self):
        return "sample_ar_fields"

    def __str__(self):
        return f"<Geoseeq::SampleResultFile {self.name} {self.uuid} />"

SampleAnalysisResultField = SampleResultFile


class ProjectResultFile(ResultFile):
    def canon_url(self):
        return "sample_group_ar_fields"

    def __str__(self):
        return f"<Geoseeq::ProjectResultFile {self.name} {self.uuid} />"

SampleGroupAnalysisResultField = ProjectResultFile



from os.path import getsize
import logging

logger = logging.getLogger("geoseeq_api")  # Same name as calling module
logger.addHandler(logging.NullHandler())


class FileChunker:

    def __init__(self, filepath, chunk_size):
        self.filepath = filepath
        self.chunk_size = chunk_size
        self.file_size = getsize(filepath)
        self.n_parts = int(self.file_size / self.chunk_size) + 1
        self.loaded_parts = []

    def load_all_chunks(self):
        if len(self.loaded_parts) != self.n_parts:
            with open(self.filepath, "rb") as f:
                f.seek(0)
                for i in range(self.n_parts):
                    chunk = f.read(self.chunk_size)
                    self.loaded_parts.append(chunk)
        return self  # convenience for chaining
    
    def chunk_is_preloaded(self, num):
        return len(self.loaded_parts) > num and self.loaded_parts[num]
    
    def read_one_chunk(self, num):
        if not self.chunk_is_preloaded(num):
            logger.debug(f"Reading chunk {num} from {self.filepath}")
            with open(self.filepath, "rb") as f:
                f.seek(num * self.chunk_size)
                chunk = f.read(self.chunk_size)
                return chunk
        return self.loaded_parts[num]

    def get_chunk(self, num):
        if self.chunk_is_preloaded(num):
            return self.loaded_parts[num]
        return self.read_one_chunk(num)
    
    def get_chunk_size(self, num):
        if num < (self.n_parts - 1):  # all but the last chunk
            return self.chunk_size
        if self.chunk_is_preloaded(num):  # last chunk, pre-loaded
            return len(self.loaded_parts[num])
        return len(self.read_one_chunk(num))  # last chunk, not pre-loaded
    



import os
import urllib
from os.path import basename, dirname, getsize, isdir, isfile, join
from pathlib import Path
from tempfile import NamedTemporaryFile

import requests

from geoseeq.constants import FIVE_MB
from geoseeq.remote_object import RemoteObject, RemoteObjectError
from geoseeq.utils import download_ftp, md5_checksum

from .bioinfo import SampleBioInfoFolder
from .result_file import ProjectResultFile, SampleResultFile
from .utils import *


class ResultFolder(RemoteObject):
    remote_fields = [
        "uuid",
        "created_at",
        "updated_at",
        "module_name",
        "replicate",
        "description",
        "is_private",
    ]

    @property
    def name(self):
        return self.module_name

    def _get(self, allow_overwrite=False):
        """Fetch the result from the server."""
        self.parent.idem()
        logger.debug(f"Getting AnalysisResult.")
        blob = self.get_cached_blob()
        if not blob:
            url = self.nested_url()
            if self.replicate:
                url += f"?replicate={self.replicate}"
            blob = self.knex.get(url, url_options=self.inherited_url_options)
            self.load_blob(blob, allow_overwrite=allow_overwrite)
            self.cache_blob(blob)
        else:
            self.load_blob(blob)

    def pre_hash(self):
        key = self.module_name + self.parent.pre_hash()
        key += self.replicate if self.replicate else ""
        return key

    def copy(self, new_parent, save=True):
        copied = new_parent.analysis_result(
            self.module_name, replicate=self.replicate, metadata=self.metadata
        )
        for field in self.get_fields():
            field.copy(copied, save=save)
        if save:
            copied.idem()
        return copied
    
    def link_file(self, link_type, file_path, remote_name=None):
        """Link a file to this result."""
        result_file = self.result_file(remote_name or basename(file_path))
        result_file.idem()
        result_file.link_file(link_type, file_path)
        return result_file
    
    def upload_file(self, file_path, remote_name=None, progress_tracker=None, chunk_size=FIVE_MB):
        """Upload a local file to GeoSeeq. Return a ResultFile object."""
        result_file = self.result_file(remote_name or basename(file_path))
        result_file.idem()
        result_file.upload_file(file_path, progress_tracker=progress_tracker, chunk_size=chunk_size)
        return result_file
    
    def _prepare_folder_upload(self, folder_path, recursive, hidden_files, prefix):
        """Yield result_file, local_path pairs for all files in a folder."""
        if isfile(folder_path):
            yield self.result_file(prefix), folder_path
            return
        for file_name in os.listdir(folder_path):
            if not hidden_files and file_name.startswith("."):
                continue
            if file_name in ["..", "."]:
                continue
            if file_name.startswith(".geoseeq") or file_name.endswith(".gs_downloaded"):
                continue
            file_path = join(folder_path, file_name)
            if isfile(file_path):
                yield self.result_file(join(prefix or "", file_name)), file_path
            elif recursive:
                yield from self._prepare_folder_upload(
                    file_path, recursive, hidden_files, join(prefix or "", file_name)
                )
    
    def upload_folder(
            self,
            folder_path,
            recursive=True,
            hidden_files=False,
            prefix=None,
            chunk_size=FIVE_MB,
            progress_tracker_factory=None,
        ):
        """Upload the contents of a local folder to geoseeq.
        
        If recursive is True, folders inside the folder will be
        uploaded as well. GeoSeeq does not create actual nested
        folders, rather files will get names that include the path
        to the file such as `folder/file.txt`.

        If hidden_files is True, files starting with a dot will be
        uploaded as well. This does not apply to `.` and `..` which
        are always ignored. Also ignoe `.geoseeq` folders.
        """
        for result_file, local_path in self._prepare_folder_upload(folder_path, recursive, hidden_files, prefix):
            result_file.upload_file(
                local_path,
                progress_tracker=progress_tracker_factory and progress_tracker_factory(local_path),
                chunk_size=chunk_size
            )

    def download_folder(self, local_folder_path, hidden_files=True):
        """Download the contents of this result folder to a local folder.
        
        If hidden_files is True, files starting with a dot will be downloaded as well.
        """
        for field in self.get_fields():
            if not hidden_files and field.name.startswith("."):
                continue
            local_file_path = join(local_folder_path, field.name)
            os.makedirs(dirname(local_file_path), exist_ok=True)
            field.download(local_file_path)
        return self

AnalysisResult = ResultFolder # for backwards compatibility


class SampleResultFolder(ResultFolder, SampleBioInfoFolder):
    parent_field = "sample"

    def __init__(self, knex, sample, module_name, replicate=None, metadata={}, is_private=False):
        super().__init__(self)
        self.knex = knex
        self.sample = sample
        self.parent = self.sample
        self.module_name = module_name
        self.replicate = replicate
        self._get_field_cache = []
        self.metadata = metadata
        self.is_private = is_private

    def nested_url(self):
        escaped_name = urllib.parse.quote(self.module_name, safe="")
        return self.sample.nested_url() + f"/analysis_results/{escaped_name}"

    def _save(self):
        data = {field: getattr(self, field) for field in self.remote_fields if hasattr(self, field)}
        data["sample"] = self.sample.uuid
        url = f"sample_ars/{self.uuid}"
        d = {"data": data, "url": url, "sample_ar": self}
        logger.debug(f"Saving SampleAnalysisResult. {d}")
        self.knex.put(url, json=data, url_options=self.inherited_url_options)

    def get_post_data(self):
        """Return a dict that can be used to POST this result to the server."""
        data = {
            field: getattr(self, field)
            for field in self.remote_fields
            if hasattr(self, field) and getattr(self, field) is not None
        }
        data["sample"] = self.sample.uuid
        if self.replicate:
            data["replicate"] = self.replicate
        return data

    def _create(self):
        self.sample.idem()
        data = self.get_post_data()
        d = {"data": data, "sample_ar": self}
        logger.debug(f"Creating SampleAnalysisResult. {d}")
        blob = self.knex.post(
            f"sample_ars?format=json", json=data, url_options=self.inherited_url_options
        )
        self.load_blob(blob)

    def result_file(self, field_name, pipeline_run=None, data={}):
        d = {
            "data": data,
            "field_name": field_name,
            "pipeline_run": pipeline_run,
            "sample_ar": self,
        }
        logger.debug(f"Creating SampleAnalysisResultField for SampleAnalysisResult. {d}")
        return SampleResultFile(self.knex, self, field_name, pipeline_run=pipeline_run, data=data)

    def field(self, *args, **kwargs):
        return self.result_file(*args, **kwargs)

    def get_result_files(self, cache=True):
        """Return a list of ar-fields fetched from the server."""
        if cache and self._get_field_cache:
            for field in self._get_field_cache:
                yield field
            return
        url = f"sample_ar_fields?analysis_result_id={self.uuid}"
        # url = self.nested_url() + f"/fields"
        logger.debug(f"Fetching SampleAnalysisResultFields. {self}")
        result = self.knex.get(url)
        for result_blob in result["results"]:
            result = self.field(result_blob["name"])
            result.load_blob(result_blob)
            # We just fetched from the server so we change the RemoteObject
            # meta properties to reflect that
            result._already_fetched = True
            result._modified = False
            if cache:
                self._get_field_cache.append(result)
            else:
                yield result
        if cache:
            for field in self._get_field_cache:
                yield field

    def get_fields(self, *args, **kwargs):
        return self.get_result_files(*args, **kwargs)

    def __str__(self):
        return f"<Geoseeq::SampleResultFolder {self.module_name} {self.replicate} {self.uuid} />"

SampleAnalysisResult = SampleResultFolder # for backwards compatibility


class ProjectResultFolder(ResultFolder):
    parent_field = "grp"

    def __init__(self, knex, grp, module_name, replicate=None, metadata={}, is_private=False):
        super().__init__(self)
        self.knex = knex
        self.grp = grp
        self.parent = self.grp
        self.module_name = module_name
        self.replicate = replicate
        self.metadata = metadata
        self.is_private = is_private

    def nested_url(self):
        return self.grp.nested_url() + f"/analysis_results/{self.module_name}"

    def _save(self):
        data = {field: getattr(self, field) for field in self.remote_fields if hasattr(self, field)}
        data["sample_group"] = self.grp.uuid
        url = f"sample_group_ars/{self.uuid}"
        self.knex.put(url, json=data)

    def _create(self):
        self.grp.idem()
        data = {
            "sample_group": self.grp.uuid,
            "module_name": self.module_name,
        }
        if self.replicate:
            data["replicate"] = self.replicate
        if self.uuid:
            data["uuid"] = self.uuid
        blob = self.knex.post(f"sample_group_ars?format=json", json=data)
        self.load_blob(blob)

    def result_file(self, field_name, pipeline_run=None, data={}):
        return ProjectResultFile(self.knex, self, field_name, pipeline_run=pipeline_run, data=data)

    def field(self, *args, **kwargs):
        return self.result_file(*args, **kwargs)

    def get_result_files(self):
        """Return a list of ar-fields fetched from the server."""
        url = f"sample_group_ar_fields?analysis_result_id={self.uuid}"
        result = self.knex.get(url)
        for result_blob in result["results"]:
            result = self.field(result_blob["name"])
            result.load_blob(result_blob)
            # We just fetched from the server so we change the RemoteObject
            # meta properties to reflect that
            result._already_fetched = True
            result._modified = False
            yield result

    def get_fields(self, *args, **kwargs):
        return self.get_result_files(*args, **kwargs)

    def __str__(self):
        return f"<Geoseeq::ProjectResultFolder {self.module_name} {self.replicate} {self.uuid} />"

SampleGroupAnalysisResult = ProjectResultFolder # for backwards compatibility
